use std::collections::HashMap;

use records::structs;
use records::types;
use common_types::Type;
use ser::RRData;

lazy_static!{
	static ref REGISTRY: Registry = Registry::init();
}

fn registry() -> &'static Registry {
	&*REGISTRY
}

pub fn known_name_to_type(name: &str) -> Option<Type> {
	let registry = registry();
	let &t = registry.names_to_type.get(name)?;
	registry.type_parser.get(&t)?;

	Some(t)
}

pub fn name_to_type(name: &str) -> Option<Type> {
	let registry = registry();
	let &t = registry.names_to_type.get(name)?;
	Some(t)
}

struct Registry {
	names_to_type: HashMap<String, Type>,
	type_parser: HashMap<Type, ()>,
	// make sure registrations are in order
	prev_type: Option<Type>,
}

impl Registry {
	fn init() -> Self {
		let mut r = Registry {
			names_to_type: HashMap::new(),
			type_parser: HashMap::new(),
			prev_type: None,
		};

		r.register_known::<structs::A>();
		r.register_known::<structs::NS>();
		r.register_known::<structs::MD>();
		r.register_known::<structs::MF>();
		r.register_known::<structs::CNAME>();
		r.register_known::<structs::SOA>();
		r.register_known::<structs::MB>();
		r.register_known::<structs::MG>();
		r.register_known::<structs::MR>();
		r.register_known::<structs::NULL>();
		r.register_known::<structs::WKS>();
		r.register_known::<structs::PTR>();
		r.register_known::<structs::HINFO>();
		r.register_known::<structs::MINFO>();
		r.register_known::<structs::MX>();
		r.register_known::<structs::TXT>();
		r.register_known::<structs::RP>();
		r.register_known::<structs::AFSDB>();
		r.register_unknown("X25"       , types::X25);
		r.register_unknown("ISDN"      , types::ISDN);
		r.register_unknown("RT"        , types::RT);
		r.register_unknown("NSAP"      , types::NSAP);
		r.register_unknown("NSAP-PTR"  , types::NSAP_PTR);
		r.register_unknown("SIG"       , types::SIG);
		r.register_known::<structs::KEY>();
		r.register_unknown("PX"        , types::PX);
		r.register_unknown("GPOS"      , types::GPOS);
		r.register_known::<structs::AAAA>();
		r.register_known::<structs::LOC>();
		r.register_unknown("NXT"       , types::NXT);
		r.register_unknown("EID"       , types::EID);
		r.register_unknown("NIMLOC"    , types::NIMLOC);
		r.register_known::<structs::SRV>();
		r.register_unknown("ATMA"      , types::ATMA);
		r.register_known::<structs::NAPTR>();
		r.register_known::<structs::KX>();
		r.register_known::<structs::CERT>();
		r.register_unknown("A6"        , types::A6);
		r.register_known::<structs::DNAME>();
		r.register_unknown("SINK"      , types::SINK);
		r.register_unknown("OPT"       , types::OPT);
		r.register_unknown("APL"       , types::APL);
		r.register_known::<structs::DS>();
		r.register_known::<structs::SSHFP>();
		r.register_unknown("IPSECKEY"  , types::IPSECKEY);
		r.register_known::<structs::RRSIG>();
		r.register_known::<structs::NSEC>();
		r.register_known::<structs::DNSKEY>();
		r.register_unknown("DHCID"     , types::DHCID);
		r.register_known::<structs::NSEC3>();
		r.register_known::<structs::NSEC3PARAM>();
		r.register_unknown("TLSA"      , types::TLSA);
		r.register_unknown("SMIMEA"    , types::SMIMEA);
		r.register_unknown("HIP"       , types::HIP);
		r.register_unknown("NINFO"     , types::NINFO);
		r.register_unknown("RKEY"      , types::RKEY);
		r.register_unknown("TALINK"    , types::TALINK);
		r.register_unknown("CDS"       , types::CDS);
		r.register_unknown("CDNSKEY"   , types::CDNSKEY);
		r.register_unknown("OPENPGPKEY", types::OPENPGPKEY);
		r.register_unknown("CSYNC"     , types::CSYNC);
		r.register_known::<structs::SPF>();
		r.register_unknown("UINFO"     , types::UINFO);
		r.register_unknown("UID"       , types::UID);
		r.register_unknown("GID"       , types::GID);
		r.register_unknown("UNSPEC"    , types::UNSPEC);
		r.register_unknown("NID"       , types::NID);
		r.register_unknown("L32"       , types::L32);
		r.register_unknown("L64"       , types::L64);
		r.register_unknown("LP"        , types::LP);
		r.register_unknown("EUI48"     , types::EUI48);
		r.register_unknown("EUI64"     , types::EUI64);
		r.register_unknown("TKEY"      , types::TKEY);
		r.register_unknown("TSIG"      , types::TSIG);
		r.register_unknown("IXFR"      , types::IXFR);
		r.register_unknown("AXFR"      , types::AXFR);
		r.register_unknown("MAILB"     , types::MAILB);
		r.register_unknown("MAILA"     , types::MAILA);
		r.register_unknown("ANY"       , types::ANY);
		r.register_unknown("URI"       , types::URI);
		r.register_unknown("CAA"       , types::CAA);
		r.register_unknown("AVC"       , types::AVC);
		r.register_unknown("DOA"       , types::DOA);
		r.register_unknown("DLV"       , types::DLV);
		r.register_unknown("ADDR"      , types::ADDR);
		r.register_unknown("ALIAS"     , types::ALIAS);

		r
	}

	fn register_unknown(&mut self, name: &'static str, rrtype: Type) {
		assert!(self.prev_type < Some(rrtype), "registration not in order");
		self.prev_type = Some(rrtype);
		assert!(self.names_to_type.insert(name.into(), rrtype).is_none());
	}

	fn register_known<T: RRData>(&mut self) {
		let n = T::rr_type_name();
		let t = T::rr_type();
		assert!(self.prev_type < Some(t), "registration not in order");
		self.prev_type = Some(t);
		assert!(self.names_to_type.insert(n.into_owned(), t).is_none());
		self.type_parser.insert(t, ());
	}
}
